package com.iot.modbus_rtcp.netty;

import cn.hutool.core.util.ArrayUtil;
import com.iot.modbus_rtcp.config.EquipmentIPProperties;
import com.iot.modbus_rtcp.dto.ModbusCommandDto;
import com.iot.modbus_rtcp.utils.HexUtil;
import com.iot.modbus_rtcp.utils.SpringUtil;
import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.ByteToMessageDecoder;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

import java.net.InetSocketAddress;
import java.util.List;
import java.util.Objects;

@Slf4j
public class ModbusDecoder extends ByteToMessageDecoder {
    private final ChannelGroup channelGroup;
    private final EquipmentIPProperties equipmentIPProperties;

    public ModbusDecoder(ChannelGroup channelGroup) {
        this.channelGroup = channelGroup;
        this.equipmentIPProperties = SpringUtil.getBean(EquipmentIPProperties.class);
    }

    @Override
    protected void decode(ChannelHandlerContext ctx, ByteBuf buffer, List<Object> out) {
        String ip = ((InetSocketAddress) ctx.channel().remoteAddress()).getHostString();
        int port = ((InetSocketAddress) ctx.channel().remoteAddress()).getPort();

        byte[] b = new byte[buffer.readableBytes()];
        buffer.readBytes(b);

        String hex = HexUtil.bytesToHexString(b);
        log.debug("解析到:{}", hex);

        String currentAddress = ChannelGroup.getKey(ip, port);
        if (this.equipmentIPProperties.contains(hex.toUpperCase())) { // 心跳
            String msgUpperCase = hex.toUpperCase();
            log.info("注册{}({}:{}):{}", this.equipmentIPProperties.get(msgUpperCase), ip, port, msgUpperCase);
            SyncPriorityChannel channel = this.channelGroup.get(currentAddress);
            if (Objects.isNull(channel)) {
                channel = this.channelGroup.getAssociatedChannel(this.equipmentIPProperties.get(msgUpperCase), msgUpperCase, ip, port);
                if (Objects.isNull(channel)) {
                    return;
                }
            }
            channel.setIdentifier(msgUpperCase);
            this.channelGroup.put(msgUpperCase, channel);

            IPGatewayRelation.put(currentAddress, String.format("%s(%s)", this.equipmentIPProperties.get(msgUpperCase), msgUpperCase));
            String oldAddress = IPGatewayRelation.getIPAddress(msgUpperCase);
            // 说明设备重连后IP+端口发生了改变
            if (!StringUtils.equals(currentAddress, oldAddress)
                    && StringUtils.isNotBlank(currentAddress) && StringUtils.isNotBlank(oldAddress)) {
                // 删除老连接
                this.channelGroup.remove(oldAddress);
                log.info("更新通道{}连接为：{}，移除旧连接：{}", this.equipmentIPProperties.get(msgUpperCase), currentAddress, oldAddress);
            }
            IPGatewayRelation.putIPAddressMap(msgUpperCase, currentAddress);
            return;
        }

        SyncPriorityChannel channel = this.channelGroup.get(currentAddress);
        ModbusCommandDto message = channel.getCurrentMessage();

        if (message == null) {
            log.warn("非法传输：{}，原报文：{}，IP：{}:{}", HexUtil.bytesToHexString(b), new String(b), ip, port);
            return;
        }

        String identity = ChannelGroup.getKey(ip, port);
        byte[] bytesCache = DataCache.get(identity);
        if (bytesCache == null) {
            bytesCache = b;
        } else {
            bytesCache = ArrayUtil.addAll(bytesCache, b);
        }

        if (bytesCache.length < (message.getLength() / 2)) {
            DataCache.put(identity, bytesCache);
        } else if (bytesCache.length == (message.getLength() / 2)) {
            out.add(HexUtil.bytesToHexString(bytesCache));
        }
    }

}
