package com.iot.modbus_rtcp.netty;

import com.iot.modbus_rtcp.config.EquipmentIPProperties;
import com.iot.modbus_rtcp.dto.ModbusCommandDto;
import com.iot.modbus_rtcp.utils.ModbusFunctionCodeEnums;
import com.iot.modbus_rtcp.utils.SpringUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 报文分发器
 * 通信报文{@link ModbusCommandDto} 中，提取命令的内容获取PLC地址，
 * 按地址将报文分发到对应的 {@link SyncPriorityChannel}
 */
@Slf4j
public class ModbusSender {
    private final ChannelGroup channelGroup;
    private final EquipmentIPProperties equipmentIPProperties;

    public ModbusSender(ChannelGroup channelGroup) {
        this.channelGroup = channelGroup;
        this.equipmentIPProperties = SpringUtil.getBean(EquipmentIPProperties.class);
    }

    public void send(List<ModbusCommandDto> modbusCommandBoList) {
        Map<String, List<ModbusCommandDto>> splitCommandMap = this.split(modbusCommandBoList);
        splitCommandMap.forEach((identity, v) -> {
            SyncPriorityChannel channel = this.channelGroup.get(identity);
            if (channel == null) {
                throw new RuntimeException("设备" + identity + "的连接未配置或该设备已断开与服务器的连接，请稍后再试或联系管理员");
            }

            log.debug("{}通道({})提交{}条命令:{}", this.equipmentIPProperties.get(identity), identity, v.size(), v);
            channel.addMessages(v);
        });
    }

    private Map<String, List<ModbusCommandDto>> split(List<ModbusCommandDto> modbusCommandBoList) {
        Map<String, List<ModbusCommandDto>> map = new HashMap<>();
        modbusCommandBoList.forEach(modbusCommandBo -> {
            if (modbusCommandBo.getIdentifier() == null) {
                throw new NullPointerException("Invalid gateway identifier");
            }
            if (!ModbusFunctionCodeEnums.isHandlerHexCode(StringUtils.substring(modbusCommandBo.getCommand(), 2, 4))) {
                throw new RuntimeException("Not support command: " + modbusCommandBo.getCommand());
            }

            List<ModbusCommandDto> list = map.getOrDefault(modbusCommandBo.getIdentifier(), new ArrayList<>());
            list.add(modbusCommandBo);

            map.put(modbusCommandBo.getIdentifier(), list);
        });

        return map;
    }
}
