package com.iot.modbus_rtcp.netty;

import com.iot.modbus.rtcp.commons.keymgr.LicenseValidator;
import com.iot.modbus_rtcp.dto.AutoCollectionDto;
import com.iot.modbus_rtcp.dto.CommandTypeComparable;
import com.iot.modbus_rtcp.dto.ModbusCommandDto;
import com.iot.modbus_rtcp.utils.HexUtil;
import com.iot.modbus_rtcp.utils.SpringUtil;
import io.netty.channel.ChannelPromise;
import io.netty.channel.socket.SocketChannel;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.QueueBuilder;
import org.springframework.amqp.rabbit.core.RabbitAdmin;

import java.net.InetSocketAddress;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.LockSupport;

/**
 * 同步控制命令优先Channel
 * 该Channel使用组合方式结合了{@link SocketChannel} 具备Channel的功能
 * 发送之后必须得到客户端响应或者超时后，执行下一条命令发送即同步发送机制，同时保证控制命令优先执行。
 * <p>
 * 调用添加命令方法作为一次任务提交给线程池执行，如果在执行期间还有调用添加命令方法便会在同一任务中追加执行的命令。
 */
@Slf4j
public class SyncPriorityChannel implements Runnable {

    private static final int allowRetryCount = 1;
    private static final boolean isControl = true;
    private final AtomicBoolean close = new AtomicBoolean(Boolean.FALSE);
    private final AtomicBoolean running = new AtomicBoolean(Boolean.FALSE);

    private final SocketChannel channel;

    private int timeoutRetryCount = 0;
    @Getter
    private volatile String identifier;
    @Getter
    private volatile String collectQueue;

    private final PriorityBlockingQueue<ModbusCommandDto> messageQueue = new PriorityBlockingQueue<>();
    private final AtomicReference<ChannelPromise> channelPromiseReference = new AtomicReference<>();
    private final AtomicReference<ModbusCommandDto> currentMessageReference = new AtomicReference<>();

    public SyncPriorityChannel(SocketChannel channel) {
        this.channel = channel;
    }

    public void addMessages(List<ModbusCommandDto> messages) {
        this.cleanAutoCollection(messages);

        this.messageQueue.addAll(messages);

        synchronized (this.running) {
            if (this.running.getAcquire()) {
                return;
            } else {
                this.running.setRelease(Boolean.TRUE);
            }
        }

        ThreadPoolConsumer.submit(this);
    }

    @Override
    public void run() {
        if (this.close.getAcquire()) {
            return;
        }
        this.sendNext();
    }

    public synchronized void sendNext() {
        if (this.timeoutRetryCount == 0) {
            this.currentMessageReference.setRelease(this.messageQueue.poll());
        }

        // 消费完毕终止递归发送
        if (Objects.isNull(this.getCurrentMessage())) {
            this.running.setRelease(Boolean.FALSE);
            return;
        }

        InetSocketAddress remoteAddress = this.channel.remoteAddress();
        int port = remoteAddress.getPort();
        String ip = remoteAddress.getHostString();

        this.channelPromiseReference.setRelease(this.channel.newPromise());
        if (isControl || Objects.equals(CommandTypeComparable.CommandType.COLLECTION, this.getCurrentMessage().getType())) {
            if (Objects.equals(CommandTypeComparable.CommandType.CONTROL, this.getCurrentMessage().getType())) {
                log.info("发送命令:{},长度:{},通道:{},IP:{},还有{}条命令待发送",
                        this.getCurrentMessage().getCommand(),
                        this.getCurrentMessage().getLength(),
                        IPGatewayRelation.get(ChannelGroup.getKey(ip, port)),
                        ChannelGroup.getKey(ip, port),
                        this.messageQueue.size());
            } else {
                log.debug("发送命令:{},长度:{},通道:{},IP:{},还有{}条命令待发送",
                        this.getCurrentMessage().getCommand(),
                        this.getCurrentMessage().getLength(),
                        IPGatewayRelation.get(ChannelGroup.getKey(ip, port)),
                        ChannelGroup.getKey(ip, port),
                        this.messageQueue.size());
            }

            this.channel.writeAndFlush(this.getCurrentMessage().getCommand());
        }

        // 清空缓存数据
        DataCache.remove(ChannelGroup.getKey(ip, port));

        boolean timeout;
        try {
            timeout = !this.getChannelPromise().await(1, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            log.error("channelPromise.await发生异常,", e);
            return;
        }

        if (timeout && this.timeoutRetryCount == (allowRetryCount - 1)) {
            log.warn("#超时命令:{},长度:{},通道:{},IP:{},还有{}条命令待发送,Command:{}",
                    this.getCurrentMessage().getCommand(),
                    this.getCurrentMessage().getLength(),
                    IPGatewayRelation.get(ChannelGroup.getKey(ip, port)),
                    ChannelGroup.getKey(ip, port),
                    this.messageQueue.size(), this.getCurrentMessage());
        }

        if (timeout && this.timeoutRetryCount < (allowRetryCount - 1)) {
            this.timeoutRetryCount++;
        } else {
            this.timeoutRetryCount = 0;
        }

        LockSupport.parkNanos(TimeUnit.MILLISECONDS.toNanos(10));

        this.sendNext();
    }

    public void close() {
        this.close.setRelease(Boolean.TRUE);
        this.messageQueue.clear();
        this.channel.close();
    }

    public ChannelPromise getChannelPromise() {
        return this.channelPromiseReference.getAcquire();
    }

    public ModbusCommandDto getCurrentMessage() {
        return this.currentMessageReference.getAcquire();
    }

    public void setIdentifier(String identifier) {
        if (StringUtils.isBlank(this.identifier)) {
            this.identifier = identifier;
            this.setCollectQueue();
        } else if (!StringUtils.equals(this.identifier, identifier)) {
            throw new RuntimeException("多个设备对应同一心跳包，当前已绑定心跳信息："
                    + new String(HexUtil.hexStringToBytes(identifier)) + "[" + identifier + "] --> "
                    + this.channel.remoteAddress().getHostString() + ":" + this.channel.remoteAddress().getPort());
        }
    }

    public void setCollectQueue() {
        this.collectQueue = "/modbus/collect/" + HexUtil.hashPartition(this.identifier, 10);
        SpringUtil.getBean(RabbitAdmin.class).declareQueue(QueueBuilder.durable(this.collectQueue).build());
    }

    private void cleanAutoCollection(List<ModbusCommandDto> messages) {
        SpringUtil.getBean(LicenseValidator.class).verify();
        if (messages.stream().anyMatch(c -> c instanceof AutoCollectionDto)) {
            this.messageQueue.removeIf(command -> command instanceof AutoCollectionDto);
        }
    }

}
